@@ -7,14 +7,19 @@ Official language server spec:
 """
 
 import itertools
-from typing import Any, List, Optional, Union
+from typing import Any, List, NamedTuple, Optional, Union
 
 import cattrs
-from jedi import Project, __version__
+from jedi import Project, Script, __version__
+from jedi.api.classes import Name
 from jedi.api.refactoring import RefactoringError
 from lsprotocol.types import (
     COMPLETION_ITEM_RESOLVE,
     INITIALIZE,
+    NOTEBOOK_DOCUMENT_DID_CHANGE,
+    NOTEBOOK_DOCUMENT_DID_CLOSE,
+    NOTEBOOK_DOCUMENT_DID_OPEN,
+    NOTEBOOK_DOCUMENT_DID_SAVE,
     TEXT_DOCUMENT_CODE_ACTION,
     TEXT_DOCUMENT_COMPLETION,
     TEXT_DOCUMENT_DECLARATION,
@@ -28,6 +33,8 @@ from lsprotocol.types import (
     TEXT_DOCUMENT_HOVER,
     TEXT_DOCUMENT_REFERENCES,
     TEXT_DOCUMENT_RENAME,
+    TEXT_DOCUMENT_SEMANTIC_TOKENS_FULL,
+    TEXT_DOCUMENT_SEMANTIC_TOKENS_RANGE,
     TEXT_DOCUMENT_SIGNATURE_HELP,
     TEXT_DOCUMENT_TYPE_DEFINITION,
     WORKSPACE_DID_CHANGE_CONFIGURATION,
@@ -41,9 +48,13 @@ from lsprotocol.types import (
     CompletionOptions,
     CompletionParams,
     DidChangeConfigurationParams,
+    DidChangeNotebookDocumentParams,
     DidChangeTextDocumentParams,
+    DidCloseNotebookDocumentParams,
     DidCloseTextDocumentParams,
+    DidOpenNotebookDocumentParams,
     DidOpenTextDocumentParams,
+    DidSaveNotebookDocumentParams,
     DidSaveTextDocumentParams,
     DocumentHighlight,
     DocumentSymbol,
@@ -55,8 +66,17 @@ from lsprotocol.types import (
     MarkupContent,
     MarkupKind,
     MessageType,
+    NotebookDocumentSyncOptions,
+    NotebookDocumentSyncOptionsNotebookSelectorType2,
+    NotebookDocumentSyncOptionsNotebookSelectorType2CellsType,
     ParameterInformation,
+    Position,
+    Range,
     RenameParams,
+    SemanticTokens,
+    SemanticTokensLegend,
+    SemanticTokensParams,
+    SemanticTokensRangeParams,
     SignatureHelp,
     SignatureHelpOptions,
     SignatureInformation,
@@ -65,11 +85,16 @@ from lsprotocol.types import (
     WorkspaceEdit,
     WorkspaceSymbolParams,
 )
+from lsprotocol.validators import INTEGER_MAX_VALUE
 from pygls.capabilities import get_capability
 from pygls.protocol import LanguageServerProtocol, lsp_method
 from pygls.server import LanguageServer
 
-from . import jedi_utils, pygls_utils, text_edit_utils
+from . import jedi_utils, notebook_utils, pygls_utils, text_edit_utils
+from .constants import (
+    SEMANTIC_TO_TOKEN_ID,
+    SUPPORTED_SEMANTIC_TYPES,
+)
 from .initialization_options import (
     InitializationOptions,
     initialization_options_converter,
@@ -131,14 +156,49 @@ class JediLanguageServerProtocol(LanguageServerProtocol):
         did_close = (
             did_close_diagnostics if diagnostics.enable else did_close_default
         )
+        did_open_notebook = (
+            did_open_notebook_diagnostics
+            if diagnostics.enable and diagnostics.did_open
+            else did_open_notebook_default
+        )
+        did_change_notebook = (
+            did_change_notebook_diagnostics
+            if diagnostics.enable and diagnostics.did_change
+            else did_change_notebook_default
+        )
+        did_save_notebook = (
+            did_save_notebook_diagnostics
+            if diagnostics.enable and diagnostics.did_save
+            else did_save_notebook_default
+        )
+        did_close_notebook = (
+            did_close_notebook_diagnostics
+            if diagnostics.enable
+            else did_close_notebook_default
+        )
         server.feature(TEXT_DOCUMENT_DID_OPEN)(did_open)
         server.feature(TEXT_DOCUMENT_DID_CHANGE)(did_change)
         server.feature(TEXT_DOCUMENT_DID_SAVE)(did_save)
         server.feature(TEXT_DOCUMENT_DID_CLOSE)(did_close)
+        server.feature(NOTEBOOK_DOCUMENT_DID_OPEN)(did_open_notebook)
+        server.feature(NOTEBOOK_DOCUMENT_DID_CHANGE)(did_change_notebook)
+        server.feature(NOTEBOOK_DOCUMENT_DID_SAVE)(did_save_notebook)
+        server.feature(NOTEBOOK_DOCUMENT_DID_CLOSE)(did_close_notebook)
 
         if server.initialization_options.hover.enable:
             server.feature(TEXT_DOCUMENT_HOVER)(hover)
 
+        if server.initialization_options.semantic_tokens.enable:
+            tokens_legend = SemanticTokensLegend(
+                token_types=SUPPORTED_SEMANTIC_TYPES, token_modifiers=[]
+            )
+            server.feature(TEXT_DOCUMENT_SEMANTIC_TOKENS_FULL, tokens_legend)(
+                semantic_tokens_full
+            )
+            server.feature(TEXT_DOCUMENT_SEMANTIC_TOKENS_RANGE, tokens_legend)(
+                semantic_tokens_range
+            )
+
         initialize_result: InitializeResult = super().lsp_initialize(params)
         workspace_options = initialization_options.workspace
         server.project = (
@@ -175,6 +235,18 @@ SERVER = JediLanguageServer(
     name="jedi-language-server",
     version=__version__,
     protocol_cls=JediLanguageServerProtocol,
+    # Advertise support for Python notebook cells.
+    notebook_document_sync=NotebookDocumentSyncOptions(
+        notebook_selector=[
+            NotebookDocumentSyncOptionsNotebookSelectorType2(
+                cells=[
+                    NotebookDocumentSyncOptionsNotebookSelectorType2CellsType(
+                        language="python"
+                    )
+                ]
+            )
+        ]
+    ),
 )
 
 
@@ -198,6 +270,7 @@ def completion_item_resolve(
         trigger_characters=[".", "'", '"'], resolve_provider=True
     ),
 )
+@notebook_utils.supports_notebooks
 def completion(
     server: JediLanguageServer, params: CompletionParams
 ) -> Optional[CompletionList]:
@@ -269,6 +342,7 @@ def completion(
     TEXT_DOCUMENT_SIGNATURE_HELP,
     SignatureHelpOptions(trigger_characters=["(", ","]),
 )
+@notebook_utils.supports_notebooks
 def signature_help(
     server: JediLanguageServer, params: TextDocumentPositionParams
 ) -> Optional[SignatureHelp]:
@@ -314,6 +388,7 @@ def signature_help(
 
 
 @SERVER.feature(TEXT_DOCUMENT_DECLARATION)
+@notebook_utils.supports_notebooks
 def declaration(
     server: JediLanguageServer, params: TextDocumentPositionParams
 ) -> Optional[List[Location]]:
@@ -331,6 +406,7 @@ def declaration(
 
 
 @SERVER.feature(TEXT_DOCUMENT_DEFINITION)
+@notebook_utils.supports_notebooks
 def definition(
     server: JediLanguageServer, params: TextDocumentPositionParams
 ) -> Optional[List[Location]]:
@@ -352,6 +428,7 @@ def definition(
 
 
 @SERVER.feature(TEXT_DOCUMENT_TYPE_DEFINITION)
+@notebook_utils.supports_notebooks
 def type_definition(
     server: JediLanguageServer, params: TextDocumentPositionParams
 ) -> Optional[List[Location]]:
@@ -397,6 +474,7 @@ def highlight(
 
 
 # Registered with HOVER dynamically
+@notebook_utils.supports_notebooks
 def hover(
     server: JediLanguageServer, params: TextDocumentPositionParams
 ) -> Optional[Hover]:
@@ -418,6 +496,7 @@ def hover(
 
 
 @SERVER.feature(TEXT_DOCUMENT_REFERENCES)
+@notebook_utils.supports_notebooks
 def references(
     server: JediLanguageServer, params: TextDocumentPositionParams
 ) -> Optional[List[Location]]:
@@ -470,7 +549,7 @@ def document_symbol(
     symbol_information = [
         symbol_info
         for symbol_info in (
-            jedi_utils.lsp_symbol_information(name)
+            jedi_utils.lsp_symbol_information(name, document.uri)
             for name in names
             if name.type != "param"
         )
@@ -534,6 +613,7 @@ def workspace_symbol(
 
 
 @SERVER.feature(TEXT_DOCUMENT_RENAME)
+@notebook_utils.supports_notebooks
 def rename(
     server: JediLanguageServer, params: RenameParams
 ) -> Optional[WorkspaceEdit]:
@@ -560,6 +640,7 @@ def rename(
         ],
     ),
 )
+@notebook_utils.supports_notebooks
 def code_action(
     server: JediLanguageServer, params: CodeActionParams
 ) -> Optional[List[CodeAction]]:
@@ -570,6 +651,13 @@ def code_action(
         2. Extract variable
         3. Extract function
     """
+    # Code actions are not yet supported for notebooks.
+    notebook = server.workspace.get_notebook_document(
+        cell_uri=params.text_document.uri
+    )
+    if notebook is not None:
+        return None
+
     document = server.workspace.get_text_document(params.text_document.uri)
     jedi_script = jedi_utils.script(server.project, document)
     code_actions = []
@@ -661,11 +749,172 @@ def did_change_configuration(
     """
 
 
+EncodedSemanticToken = NamedTuple(
+    "EncodedSemanticToken",
+    [
+        ("line", int),
+        ("start", int),
+        ("length", int),
+        ("tokenType", int),
+        ("tokenModifiers", int),
+    ],
+)
+"""
+Semantic token encoded into integers. Applicable for both absolute and relative positions.
+
+See the LSP spec for details:
+    <https://microsoft.github.io/language-server-protocol/specifications/lsp/3.17/specification/#textDocument_semanticTokens>
+"""
+
+
+def _raw_semantic_token(
+    server: JediLanguageServer, n: Name
+) -> Union[EncodedSemanticToken, None]:
+    """Find an appropriate semantic token for the name.
+
+    This works by looking up the definition (using jedi ``goto``) of the name and
+    matching the definition's type to one of the availabile semantic tokens. Further
+    improvements are possible by inspecting context, e.g. semantic token modifiers such
+    as ``abstract`` or ``async`` or even different tokens, e.g. ``property`` or
+    ``method``. Dunder methods may warrant special treatment/modifiers as well.
+    The return is a "raw" semantic token rather than a "diff." This is in the form of a
+    length 5 array of integers where the elements are the line number, starting
+    character, length, token index, and modifiers (as an integer whose binary
+    representation has bits set at the indices of all applicable modifiers).
+    """
+    definitions: list[Name] = n.goto(
+        follow_imports=True,
+        follow_builtin_imports=True,
+        only_stubs=False,
+        prefer_stubs=False,
+    )
+    if not definitions:
+        server.show_message_log(
+            f"no definitions found for name \"{n.description}\" of type '{n.type}' ({n.line}:{n.column})",
+            MessageType.Debug,
+        )
+        return None
+
+    if len(definitions) > 1:
+        def_lines = "\n".join(
+            map(lambda n: str(n), definitions)
+        )  # f-string expression part cannot include a backslash
+        msg = (
+            f"multiple definitions found for name \"{n.description}\" of type '{n.type}' ({n.line}:{n.column}):\n"
+            f" {def_lines}"
+        )
+        server.show_message_log(msg, MessageType.Debug)
+
+    definition, *_ = definitions
+    definition_type = SEMANTIC_TO_TOKEN_ID.get(definition.type, None)
+    if definition_type is None:
+        server.show_message_log(
+            f"no matching semantic token for \"{n.description}\" of type '{n.type}' ({n.line}:{n.column})",
+            MessageType.Debug,
+        )
+        return None
+
+    return EncodedSemanticToken(
+        n.line - 1, n.column, len(n.name), definition_type, 0
+    )
+
+
+@SERVER.thread()
+def semantic_tokens_full(
+    server: JediLanguageServer, params: SemanticTokensParams
+) -> SemanticTokens:
+    """Thin wrap around  _semantic_tokens_range()."""
+    document = server.workspace.get_text_document(params.text_document.uri)
+    jedi_script = jedi_utils.script(server.project, document)
+
+    server.show_message_log(
+        f"semantic_tokens_full {params.text_document.uri} ",
+        MessageType.Log,
+    )
+
+    return _semantic_tokens_range(
+        server,
+        jedi_script,
+        Range(Position(0, 0), Position(INTEGER_MAX_VALUE, INTEGER_MAX_VALUE)),
+    )
+
+
+@SERVER.thread()
+def semantic_tokens_range(
+    server: JediLanguageServer, params: SemanticTokensRangeParams
+) -> SemanticTokens:
+    """Thin wrap around  _semantic_tokens_range()."""
+    document = server.workspace.get_text_document(params.text_document.uri)
+    jedi_script = jedi_utils.script(server.project, document)
+
+    server.show_message_log(
+        f"semantic_tokens_range {params.text_document.uri} {params.range}",
+        MessageType.Log,
+    )
+
+    return _semantic_tokens_range(server, jedi_script, params.range)
+
+
+def _semantic_tokens_range(
+    server: JediLanguageServer, jedi_script: Script, doc_range: Range
+) -> SemanticTokens:
+    """General purpose function to do full / range semantic tokens."""
+    line, column = doc_range.start.line, doc_range.start.character
+    names = jedi_script.get_names(
+        all_scopes=True, definitions=True, references=True
+    )
+    data: list[int] = []
+
+    for n in names:
+        if (
+            not doc_range.start
+            < Position(n.line - 1, n.column)
+            < doc_range.end
+        ):
+            continue
+
+        token = _raw_semantic_token(server, n)
+
+        # server.show_message_log(
+        #     f"raw token for name {n.description} ({n.line - 1}:{n.column}): {token}",
+        #     MessageType.Debug,
+        # )
+        if token is None:
+            continue
+
+        token_line, token_column = token.line, token.start
+        delta_column = (
+            token_column - column if token_line == line else token_column
+        )
+        delta_line = token_line - line
+
+        line = token_line
+        column = token_column
+
+        # server.show_message_log(
+        #     f"diff token for name {n.description} ({n.line - 1}:{n.column}): {token}",
+        #     MessageType.Debug,
+        # )
+        data.extend(
+            [
+                delta_line,
+                delta_column,
+                token.length,
+                token.tokenType,
+                token.tokenModifiers,
+            ]
+        )
+
+    return SemanticTokens(data=data)
+
+
 # Static capability or initializeOptions functions that rely on a specific
 # client capability or user configuration. These are associated with
 # JediLanguageServer within JediLanguageServerProtocol.lsp_initialize
 @jedi_utils.debounce(1, keyed_by="uri")
-def _publish_diagnostics(server: JediLanguageServer, uri: str) -> None:
+def _publish_diagnostics(
+    server: JediLanguageServer, uri: str, filename: Optional[str] = None
+) -> None:
     """Helper function to publish diagnostics for a file."""
     # The debounce decorator delays the execution by 1 second
     # canceling notifications that happen in that interval.
@@ -673,9 +922,11 @@ def _publish_diagnostics(server: JediLanguageServer, uri: str) -> None:
     # whether the document still exists
     if uri not in server.workspace.documents:
         return
+    if filename is None:
+        filename = uri
 
     doc = server.workspace.get_text_document(uri)
-    diagnostic = jedi_utils.lsp_python_diagnostic(uri, doc.source)
+    diagnostic = jedi_utils.lsp_python_diagnostic(filename, doc.source)
     diagnostics = [diagnostic] if diagnostic else []
 
     server.publish_diagnostics(uri, diagnostics)
@@ -731,7 +982,7 @@ def did_close_diagnostics(
     server: JediLanguageServer, params: DidCloseTextDocumentParams
 ) -> None:
     """Actions run on textDocument/didClose: diagnostics."""
-    server.publish_diagnostics(params.text_document.uri, [])
+    _clear_diagnostics(server, params.text_document.uri)
 
 
 def did_close_default(
@@ -741,6 +992,106 @@ def did_close_default(
     """Actions run on textDocument/didClose: default."""
 
 
+# NOTEBOOK_DOCUMENT_DID_SAVE
+def did_save_notebook_diagnostics(
+    server: JediLanguageServer,
+    params: DidSaveNotebookDocumentParams,
+) -> None:
+    """Actions run on notebookDocument/didSave: diagnostics."""
+    notebook_document = server.workspace.get_notebook_document(
+        notebook_uri=params.notebook_document.uri
+    )
+    if notebook_document:
+        for cell in notebook_document.cells:
+            text_document = server.workspace.text_documents[cell.document]
+            _publish_cell_diagnostics(server, text_document.uri)
+
+
+def did_save_notebook_default(
+    server: JediLanguageServer,
+    params: DidSaveNotebookDocumentParams,
+) -> None:
+    """Actions run on notebookDocument/didSave: default."""
+
+
+# NOTEBOOK_DOCUMENT_DID_CHANGE
+def did_change_notebook_diagnostics(
+    server: JediLanguageServer,
+    params: DidChangeNotebookDocumentParams,
+) -> None:
+    """Actions run on notebookDocument/didChange: diagnostics."""
+    cells = params.change.cells
+    if cells:
+        structure = cells.structure
+        if structure:
+            did_open = structure.did_open
+            if did_open:
+                for text_document_item in did_open:
+                    _publish_cell_diagnostics(
+                        server,
+                        text_document_item.uri,
+                    )
+            did_close = structure.did_close
+            if did_close:
+                for text_document in did_close:
+                    _clear_diagnostics(server, text_document.uri)
+        text_content = cells.text_content
+        if text_content:
+            for change in text_content:
+                _publish_cell_diagnostics(server, change.document.uri)
+
+
+def did_change_notebook_default(
+    server: JediLanguageServer,
+    params: DidChangeNotebookDocumentParams,
+) -> None:
+    """Actions run on notebookDocument/didChange: default."""
+
+
+# NOTEBOOK_DOCUMENT_DID_OPEN
+def did_open_notebook_diagnostics(
+    server: JediLanguageServer,
+    params: DidOpenNotebookDocumentParams,
+) -> None:
+    """Actions run on notebookDocument/didOpen: diagnostics."""
+    for text_document in params.cell_text_documents:
+        _publish_cell_diagnostics(server, text_document.uri)
+
+
+def did_open_notebook_default(
+    server: JediLanguageServer,
+    params: DidOpenNotebookDocumentParams,
+) -> None:
+    """Actions run on notebookDocument/didOpen: default."""
+
+
+# NOTEBOOK_DOCUMENT_DID_CLOSE
+def did_close_notebook_diagnostics(
+    server: JediLanguageServer,
+    params: DidCloseNotebookDocumentParams,
+) -> None:
+    """Actions run on notebookDocument/didClose: diagnostics."""
+    for text_document in params.cell_text_documents:
+        _clear_diagnostics(server, text_document.uri)
+
+
+def did_close_notebook_default(
+    server: JediLanguageServer,
+    params: DidCloseNotebookDocumentParams,
+) -> None:
+    """Actions run on notebookDocument/didClose: default."""
+
+
+def _clear_diagnostics(server: JediLanguageServer, uri: str) -> None:
+    """Helper function to clear diagnostics for a file."""
+    server.publish_diagnostics(uri, [])
+
+
+def _publish_cell_diagnostics(server: JediLanguageServer, uri: str) -> None:
+    filename = notebook_utils.cell_filename(server.workspace, uri)
+    return _publish_diagnostics(server, uri, filename)
+
+
 def _choose_markup(server: JediLanguageServer) -> MarkupKind:
     """Returns the preferred or first of supported markup kinds."""
     markup_preferred = server.initialization_options.markup_kind_preferred
